import pytest
import numpy as np
from scipy.spatial.distance import jaccard
from rdkit.Chem import AllChem
from mhfp.encoder import MHFPEncoder
from mhfp.lsh_forest import LSHForestHelper

# Keeping tests barebone and simple

mhfp_encoder = MHFPEncoder()
lfh = LSHForestHelper()

drugbank = []

with open('test/drugbank.smi') as f:
    for line in f.readlines():
        mol = AllChem.MolFromSmiles(line.strip().split()[0])
        if mol:
            drugbank.append(mhfp_encoder.encode_mol(mol))

for i, fp in enumerate(drugbank):
    lfh.add(i, fp)

lfh.index()

def test_setup():
    assert len(drugbank) == 226

def test_add():
    assert len(lfh.lsh_forest.keys) == 226

def test_query():
    assert lfh.query(drugbank[0], 6, drugbank) == [0, 3, 4, 7, 5, 9]
    assert lfh.query(drugbank[0], 6, drugbank, 100) == [0, 3, 4, 7, 5, 9]

def test_query_exclude():
    assert lfh.query_exclude(drugbank[0], 6, drugbank, [0, 1, 2, 3, 4, 5]) == [7, 9, 171, 20, 30, 62]






import warnings
import numpy as np
from functools import partial
from operator import itemgetter
from collections import defaultdict
from mhfp.encoder import MHFPEncoder

# Adapted from:
# https://github.com/ekzhu/datasketch/blob/master/datasketch/lshforest.py  

class LSHForestHelper():
  """A class simplyfing the use of the LSH Forest implementation. Adding paramter kc to imporve search results.
  """
  def __init__(self, dims = 2048, n_prefix_trees = 64, memmap=False, path=''):
    """The parameter dims has to be set to the same number with which the MHFP fingerprints were encoded.

    Keyword arguments:
        dims {int} -- The number of dimensions the MHFP fingerprints to be indexed were encoded in (default: {2048})
        n_prefix_trees {int} -- The number of prefix trees to use for indexing (default: {64})

    """

    self.dims = dims
    self.n_prefix_trees = n_prefix_trees
    self.memmap = memmap
    self.path = path

    self.lsh_forest = LSHForest(self.dims, n_prefix_trees=self.n_prefix_trees)

  def add(self, key, mhfp):
    """Adds a MHFP stored in a numpy.ndarray to the LSH Forest. The index() method has to be run before
    entries can be queried.

    Keyword arguments:
        key -- The key for the MHFP entry.
        mhfp {numpy.ndarray} -- A MHFP array containing 32-bit hashes
    """

    self.lsh_forest.add(key, mhfp)

  def index(self):
    """Indexes the added entities. This method has to be run after all MHFPs have been added or when
    new ones were added. Has to be run before querying.
    """

    self.lsh_forest.index()

  def query(self, query_mhfp, k, data, kc=10):
    """ Query the LSH Forest. k is multiplied by kc when querying the LSH Forest. The k nearest neighbors
    ist then selected from the list of k * kc nearest neighbors using linear scan.

    Keyword arguments:
        query_mhfp {numpy.ndarray} -- The query MHFP fingerprint.
        k {int} -- The number of nearest neighbors to be returned from the approximate nearest neighbors
        data {dict} -- The MHFP values indexed with the same key supplied to add()
        kc {int} -- The factor in kc * k for additional approximate nearest neighbors to be searched for (default: {10})
    Returns:
      numpy.ndarray -- An array containing the merged hash values.
    """

    results = self.lsh_forest.query(query_mhfp, k * kc)
    return LSHForestHelper._get_knn(query_mhfp, results, k, data)

  def query_exclude(self, query_mhfp, k, data, exclude, kc=10):
    """ Query the LSH Forest. k is multiplied by kc when querying the LSH Forest. The k nearest neighbors
    ist then selected from the list of k * kc nearest neighbors using linear scan.

    Keyword arguments:
        query_mhfp {numpy.ndarray} -- The query MHFP fingerprint.
        k {int} -- The number of nearest neighbors to be returned from the approximate nearest neighbors
        data {dict} -- The MHFP values indexed with the same key supplied to add()
        exclude {list} -- An array of ids to be ignored
        kc {int} -- The factor in kc * k for additional approximate nearest neighbors to be searched for (default: {10})
    Returns:
      numpy.ndarray -- An array containing the merged hash values.
    """

    results = self.lsh_forest.query_exclude(query_mhfp, k * kc, exclude)
    return LSHForestHelper._get_knn(query_mhfp, results, k, data)

  @staticmethod
  def _get_knn(query_mhfp, ann, k, data):
    """ Brute-force search for selecting k nearest neighbors from k * kc  approximate nearest neighbors.

    Keyword arguments:
        query_mhfp {numpy.ndarray} -- The query MHFP fingerprint.
        ann {list} -- A list of indices of approximate nearest neighbors of size k * kc to be brute-force searched
        k {int} -- The number of nearest neighbors to be returned from the approximate nearest neighbors
        data {dict} -- The MHFP values indexed with the same key supplied to add()
    """

    dists = []

    for index in ann:
        dists.append((index, 1.0 - MHFPEncoder.distance(query_mhfp, data[index])))
    
    dists.sort(key=itemgetter(1), reverse=True)
    return [x[0] for x in dists[:k]]


class LSHForest():
  """A class for indexing and querying MHFP fingerprints using the LSH Forest algorithm.
  """

  def __init__(self, dims = 2048, n_prefix_trees = 64):
    """The parameter dims has to be set to the same number with which the MHFP fingerprints were encoded.

    Keyword arguments:
        dims {int} -- The number of dimensions the MHFP fingerprints to be indexed were encoded in (default: {2048})
        n_prefix_trees {int} -- The number of prefix trees to use for indexing (default: {64})
    """
    self.count = 0
    self.dims = dims
    self.n_prefix_trees = n_prefix_trees

    self.max_depth = int(self.dims / self.n_prefix_trees)
    self.hashtables = [defaultdict(list) for _ in range(self.n_prefix_trees)]
    self.hashranges = np.array([np.array([i * self.max_depth, (i + 1) * self.max_depth]) for i in range(self.n_prefix_trees)])
    self.keys = dict()
    self.clean = False

    self.hashtables_sorted = [[] for _ in range(self.n_prefix_trees)]

  def add(self, key, mhfp):
    """Adds a MHFP stored in a numpy.ndarray to the LSH Forest. The index() method has to be run before
    entries can be queried.

    Keyword arguments:
        key {int} -- The key for the MHFP entry.
        mhfp {numpy.ndarray} -- A MHFP array containing 32-bit hashes
    """

    # self.keys[key] = [self._swap(mhfp[start:end]) for start, end in self.hashranges]
    self.keys[key] = 0
    
    k = [self._swap(mhfp[start:end]) for start, end in self.hashranges]

    for h, hashtable in zip(k, self.hashtables):
      hashtable[h].append(key)
    
    self.clean = False

  def index(self):
    """Indexes the added entities. This method has to be run after all MHFPs have been added or when
    new ones were added. Has to be run before querying.
    """
    for i, hashtable in enumerate(self.hashtables):
      self.hashtables_sorted[i] = np.array([k for k in hashtable.keys()])
      self.hashtables_sorted[i].sort()
    
    self.clean = True

  def query(self, mhfp, k):
    """Query the LSH Forest.

    Keyword arguments:
        mhfp -- The query MHFP fingerprint
        k -- The number of nearest neighbors to be searched for
    """
    if not self.clean:
      warnings.warn('Index out of date. Call index() before querying after adding new entities.', UserWarning)

    results = set()
    r = self.max_depth

    while r > 0:
      for key in self._internal_query(mhfp, r):
        results.add(key)
        if len(results) >= k:
          print(self.count)
          return list(results)
      r -= 1
    return list(results)

  def query_exclude(self, mhfp, k, exclude):
    """Query the LSH Forest.

    Keyword arguments:
        mhfp -- The query MHFP fingerprint
        k -- The number of nearest neighbors to be searched for
        exclude -- An array of ids to be ignored
    """
    if not self.clean:
      warnings.warn('Index out of date. Call index() before querying after adding new entities.', UserWarning)

    results = set()
    r = self.max_depth

    while r > 0:
      for key in self._internal_query(mhfp, r):
        if key not in exclude:
          results.add(key)
        if len(results) >= k:
          return list(results)
      r -= 1
    return list(results)

  def _internal_query(self, mhfp, r):
    """Internal method, use the query method instead.

    Keyword arguments:
        mhfp -- The query MHFP fingerprint
        r -- The current search depth
    """

    prefixes = [self._swap(mhfp[start:start + r]) for start, _ in self.hashranges]
    len_prefix = len(prefixes[0])

    for hashtable_sorted, hashtable, prefix in zip(self.hashtables_sorted, self.hashtables, prefixes):
      i = self._binary_search(len(hashtable_sorted), hashtable_sorted, prefix, len_prefix)
      
      if i < len(hashtable_sorted) and hashtable_sorted[i][:len_prefix] == prefix:
        j = i
        while j < len(hashtable_sorted) and hashtable_sorted[j][:len_prefix] == prefix:
          for key in hashtable[hashtable_sorted[j]]:
            yield key
          j += 1

  def _swap(self, hashes):
    """Internal method. Swaps bytes for prefixing.

    Keyword arguments:
        hashes -- The hashes on which to perform a byteswap
    """
    return bytes(hashes.byteswap().data)


  def _binary_search(self, n, hashtable_sorted, prefix, len_prefix):
    """Internal method. Performs a binary search on a hashtable using func as the comparison function.

    Keyword arguments:
        n -- Upper bound for index
        hashtable_sorted -- A sorted hashtable
        prefix -- The prefix from the query vector
        len_prefix -- The length of the input prefix
    """

    i = 0
    j = n
    while i < j:
      h = int(i + (j - i) / 2)

      if not hashtable_sorted[h][:len_prefix] >= prefix:
        i = h + 1
      else:
        j = h

    return i


import struct
import numpy as np

from hashlib import sha1
from rdkit.Chem import AllChem


class MHFPEncoder:
    """A class for encoding SMILES and RDKit molecule instances as MHFP fingerprints.
  """

    prime = (1 << 61) - 1
    max_hash = (1 << 32) - 1

    def __init__(self, n_permutations=2048, seed=42):
        """All minhashes created using this instance will use the arguments
    supplied to the constructor.

    Keyword arguments:
        n_permutations {int} -- The number of permutations used for hashing (default: {128})
        seed {int} -- The value used to seed numpy.random (default: {42})
    """
        self.n_permutations = n_permutations
        self.seed = seed

        self.permutations_a = np.zeros([n_permutations], dtype=np.uint32)
        self.permutations_b = np.zeros([n_permutations], dtype=np.uint32)

        self.permutations_a_64 = np.zeros([n_permutations], dtype=np.uint64)
        self.permutations_b_64 = np.zeros([n_permutations], dtype=np.uint64)

        rand = np.random.RandomState(self.seed)

        # This is done in a loop as there shouldn't be any duplicate random numbers.
        # Also, numpy.random.choice seems to be implemented badly, as it throws
        # a memory error when supplied with a large n.
        for i in range(n_permutations):
            a = rand.randint(1, MHFPEncoder.max_hash, dtype=np.uint32)
            b = rand.randint(0, MHFPEncoder.max_hash, dtype=np.uint32)

            while a in self.permutations_a:
                a = rand.randint(1, MHFPEncoder.max_hash, dtype=np.uint32)

            while b in self.permutations_b:
                b = rand.randint(0, MHFPEncoder.max_hash, dtype=np.uint32)

            self.permutations_a[i] = a
            self.permutations_b[i] = b

        # Reshape into column vectors
        self.permutations_a = self.permutations_a.reshape((self.n_permutations, 1))
        self.permutations_b = self.permutations_b.reshape((self.n_permutations, 1))

    def encode(
        self,
        in_smiles,
        radius=3,
        rings=True,
        kekulize=True,
        min_radius=1,
        sanitize=True,
    ):
        """Creates an MHFP array from a SMILES string.
    
    Arguments:
      in_smiles {string} -- A valid SMILES string
      radius {int} -- The MHFP radius (a radius of 3 corresponds to MHFP6)  (default: {3})
      rings {boolean} -- Whether or not to include rings in the shingling (default: {True})
      kekulize {boolean} -- Whether or not to kekulize the extracted SMILES (default: {True})
      sanitize {boolean} -- Whether or not to sanitize the SMILES when parsing it using RDKit  (default: {True})
      min_radius {int} -- The minimum radius that is used to extract n-grams (default: {1})
    
    Returns:
      numpy.ndarray -- An array containing the MHFP hash values.
    """

        return self.from_molecular_shingling(
            self.shingling_from_smiles(
                in_smiles,
                radius=radius,
                rings=rings,
                kekulize=kekulize,
                sanitize=sanitize,
                min_radius=min_radius,
            )
        )

    def encode_mol(self, in_mol, radius=3, rings=True, kekulize=True, min_radius=1):
        """ Creates an MHFP array from an RDKit molecule.
    
    Arguments:
      in_mol {rdkit.Chem.rdchem.Mol} -- A RDKit molecule instance
      radius {int} -- The MHFP radius (a radius of 3 corresponds to MHFP6)  (default: {3})
      rings {boolean} -- Whether or not to include rings in the shingling (default: {True})
      kekulize {boolean} -- Whether or not to kekulize the the extracted SMILES (default: {True})
      min_radius {int} -- The minimum radius that is used to extract n-grams (default: {1})
    
    Returns:
      numpy.ndarray -- An array containing the MHFP hash values.
    """

        return self.from_molecular_shingling(
            self.shingling_from_mol(
                in_mol,
                radius=radius,
                rings=rings,
                kekulize=kekulize,
                min_radius=min_radius,
            )
        )

    def from_molecular_shingling(self, tokens):
        """Creates the hash set for a string array and returns it without changing the hash values of
    this instance.
    
    Arguments:
      a {numpy.ndarray} -- A string array.
    
    Returns:
      numpy.ndarray -- An array containing the hash values.
    """

        hash_values = np.zeros([self.n_permutations, 1], dtype=np.uint32)
        hash_values.fill(MHFPEncoder.max_hash)

        for t in tokens:
            t_h = struct.unpack("<I", sha1(t).digest()[:4])[0]
            hashes = np.remainder(
                np.remainder(
                    self.permutations_a * t_h + self.permutations_b, MHFPEncoder.prime
                ),
                self.max_hash,
            )
            hash_values = np.minimum(hash_values, hashes)

        return hash_values.reshape((1, self.n_permutations))[0]

    def from_sparse_array(self, array):
        """Creates the hash set for a sparse binary array and returns it without changing the hash
    values of this instance. This is useful when a molecular shingling is already hashed.
    
    Arguments:
      s {numpy.ndarray} -- A sparse binary array.
    
    Returns:
      numpy.ndarray -- An array containing the hash values.
    """

        hash_values = np.zeros([self.n_permutations, 1], dtype=np.uint32)
        hash_values.fill(MHFPEncoder.max_hash)

        for i in array:
            hashes = np.remainder(
                np.remainder(
                    self.permutations_a * i + self.permutations_b, MHFPEncoder.prime
                ),
                self.max_hash,
            )
            hash_values = np.minimum(hash_values, hashes)

        return hash_values.reshape((1, self.n_permutations))[0]

    def from_binary_array(self, array):
        """Creates the hash set for a binary array and returns it without changing the hash
    values of this instance. This is useful to minhash a folded fingerprint.
    
    Arguments:
      s {numpy.ndarray} -- A sparse binary array.
    
    Returns:
      numpy.ndarray -- A binary array.
    """

        hash_values = np.zeros([self.n_permutations, 1], dtype=np.uint32)
        hash_values.fill(MHFPEncoder.max_hash)

        for i, v in enumerate(array):
            if v == 1:
                hashes = np.remainder(
                    np.remainder(
                        self.permutations_a * i + self.permutations_b, MHFPEncoder.prime
                    ),
                    self.max_hash,
                )
                hash_values = np.minimum(hash_values, hashes)

        return hash_values.reshape((1, self.n_permutations))[0]

    @staticmethod
    def hash(shingling):
        """ For testing purposes only. """

        hash_values = []

        for t in shingling:
            hash_values.append(struct.unpack("<I", sha1(t).digest()[:4])[0])

        return np.array(hash_values)

    @staticmethod
    def merge(a, b):
        """Merges (union) the two MHFP vectors.
    
    Arguments:
      a {numpy.ndarray} -- An array containing hash values.
      b {numpy.ndarray} -- An array containing hash values.
    Returns:
      numpy.ndarray -- An array containing the merged hash values.
    """
        return np.minimum(a, b)

    @staticmethod
    def merge_all(hash_values):
        """Merges (union) a list of hash_values.
    
    Arguments:
      hash_values {numpy.ndarray} -- An array of lists or arrays containing hash values.
    Returns:
      numpy.ndarray -- An array containing the merged hash values.
    """
        return np.amin(hash_values)

    @staticmethod
    def distance(a, b):
        """Estimates the Jaccard distance of two binary arrays based on their hashes.
    
    Arguments:
      a {numpy.ndarray} -- An array containing hash values.
      b {numpy.ndarray} -- An array containing hash values.
    
    Returns:
      float -- The estimated Jaccard distance.
    """

        # The Jaccard distance of Minhashed values is estimated by
        return 1.0 - np.float(np.count_nonzero(a == b)) / np.float(len(a))

    @staticmethod
    def shingling_from_mol(in_mol, radius=3, rings=True, kekulize=True, min_radius=1):
        """Creates a molecular shingling from a RDKit molecule (rdkit.Chem.rdchem.Mol).
    
    Arguments:
      in_mol {rdkit.Chem.rdchem.Mol} -- A RDKit molecule instance
      radius {int} -- The MHFP radius (a radius of 3 corresponds to MHFP6)  (default: {3})
      rings {boolean} -- Whether or not to include rings in the shingling (default: {True})
      kekulize {boolean} -- Whether or not to kekulize the the extracted SMILES (default: {True})
      min_radius {int} -- The minimum radius that is used to extract n-grams (default: {1})
    
    Returns:
      list -- The molecular shingling.
    """

        shingling = []

        if rings:
            for ring in AllChem.GetSymmSSSR(in_mol):
                bonds = set()
                ring = list(ring)
                for i in ring:
                    for j in ring:
                        if i != j:
                            bond = in_mol.GetBondBetweenAtoms(i, j)
                            if bond != None:
                                bonds.add(bond.GetIdx())
                shingling.append(
                    AllChem.MolToSmiles(
                        AllChem.PathToSubmol(in_mol, list(bonds)), canonical=True
                    ).encode("utf-8")
                )

        if min_radius == 0:
            for i, atom in enumerate(in_mol.GetAtoms()):
                shingling.append(atom.GetSmarts().encode("utf-8"))

        for index, _ in enumerate(in_mol.GetAtoms()):
            for i in range(1, radius + 1):
                p = AllChem.FindAtomEnvironmentOfRadiusN(in_mol, i, index)
                amap = {}
                submol = AllChem.PathToSubmol(in_mol, p, atomMap=amap)

                if index not in amap:
                    continue

                smiles = AllChem.MolToSmiles(
                    submol, rootedAtAtom=amap[index], canonical=True
                )

                if smiles is not "":
                    shingling.append(smiles.encode("utf-8"))

        # Set ensures that the same shingle is not hashed multiple times
        # (which would not change the hash, since there would be no new minima)
        return list(set(shingling))

    @staticmethod
    def shingling_from_smiles(
        in_smiles, radius=3, rings=True, kekulize=True, min_radius=1, sanitize=False
    ):
        """Creates a molecular shingling from a SMILES string.
    
    Arguments:
      in_smiles {string} -- A valid SMILES string
      radius {int} -- The MHFP radius (a radius of 3 corresponds to MHFP6)  (default: {3})
      rings {boolean} -- Whether or not to include rings in the shingling (default: {True})
      kekulize {boolean} -- Whether or not to kekulize the extracted SMILES (default: {True})
      min_radius {int} -- The minimum radius that is used to extract n-grams (default: {1})
      sanitize {boolean} -- Whether or not to sanitize the SMILES when parsing it using RDKit  (default: {False})
    
    Returns:
      list -- The molecular shingling.
    """

        return MHFPEncoder.shingling_from_mol(
            AllChem.MolFromSmiles(in_smiles, sanitize=sanitize),
            rings=rings,
            radius=radius,
            kekulize=True,
            min_radius=min_radius,
        )

    @staticmethod
    def fold(hash_values, length=2048):
        """Folds the hash values to a binary vector of a given length.

    Arguments:
      hash_value {numpy.ndarray} -- An array containing the hash values.
      length {int} -- The length of the folded fingerprint (default: {2048})
    
    Returns:
      numpy.ndarray -- The folded fingerprint.
    """

        folded = np.zeros(length, dtype=np.uint8)
        folded[hash_values % length] = 1

        return folded

    @staticmethod
    def secfp_from_mol(
        in_mol, length=2048, radius=3, rings=True, kekulize=True, min_radius=1
    ):
        """Creates a folded binary vector fingerprint of a input molecule.

    Arguments:
      in_mol {rdkit.Chem.rdchem.Mol} -- A RDKit molecule instance
      length {int} -- The length of the folded fingerprint (default: {2048})
      radius {int} -- The MHFP radius (a radius of 3 corresponds to SECFP6)  (default: {3})
      rings {boolean} -- Whether or not to include rings in the shingling (default: {True})
      kekulize {boolean} -- Whether or not to kekulize the extracted SMILES (default: {True})
      min_radius {int} -- The minimum radius that is used to extract n-grams (default: {1})
    
    Returns:
      numpy.ndarray -- The folded fingerprint.
    """
        return MHFPEncoder.fold(
            MHFPEncoder.hash(
                MHFPEncoder.shingling_from_mol(
                    in_mol,
                    radius=radius,
                    rings=rings,
                    kekulize=kekulize,
                    min_radius=min_radius,
                )
            ),
            length=length,
        )

    @staticmethod
    def secfp_from_smiles(
        in_smiles, length=2048, radius=3, rings=True, kekulize=True, sanitize=False
    ):
        """Creates a folded binary vector fingerprint of a input SMILES string.

    Arguments:
      in_smiles {string} -- A valid SMILES string
      length {int} -- The length of the folded fingerprint (default: {2048})
      radius {int} -- The MHFP radius (a radius of 3 corresponds to SECFP6)  (default: {3})
      rings {boolean} -- Whether or not to include rings in the shingling (default: {True})
      kekulize {boolean} -- Whether or not to kekulize the extracted SMILES (default: {True})
      sanitize {boolean} -- Whether or not to sanitize the SMILES when parsing it using RDKit  (default: {False})
    
    Returns:
      numpy.ndarray -- The folded fingerprint.
    """
        return MHFPEncoder.secfp_from_mol(
            AllChem.MolFromSmiles(in_smiles, sanitize=sanitize),
            length=length,
            radius=radius,
            rings=rings,
            kekulize=kekulize,
        )




import pytest
import numpy as np
from scipy.spatial.distance import jaccard
from rdkit.Chem import AllChem
from mhfp.encoder import MHFPEncoder
from mhfp.lsh_forest import LSHForest

# Keeping tests barebone and simple

mhfp_encoder = MHFPEncoder()
lf = LSHForest()

drugbank = []

with open('test/drugbank.smi') as f:
    for line in f.readlines():
        mol = AllChem.MolFromSmiles(line.strip().split()[0])
        if mol:
            drugbank.append(mhfp_encoder.encode_mol(mol))

for i, fp in enumerate(drugbank):
    lf.add(i, fp)

lf.index()

def test_setup():
    assert len(drugbank) == 226

def test_add():
    assert len(lf.keys) == 226

def test_query():
    assert lf.query(drugbank[0], 6) == [0, 3, 4, 197, 9, 149]

def test_query_exclude():
    assert lf.query_exclude(drugbank[0], 4, [0, 9]) == [197, 3, 4, 149]




import sys
import os
from setuptools import setup
from setuptools import Command
from setuptools.command.test import test as TestCommand
from datetime import datetime
import mhfp


with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

NAME = "mhfp"
VERSION = mhfp.__version__
AUTHOR = "Daniel Probst"
DESCRIPTION = "Molecular MHFP fingerprints for cheminformatics applications"
URL = "https://github.com/reymond-group/mhfp"
REQUIRED_PYTHON_VERSION = (3, 0)
PACKAGES = ["mhfp"]
INSTALL_DEPENDENCIES = []
SETUP_DEPENDENCIES = []
TEST_DEPENDENCIES = ["pytest"]
EXTRA_DEPENDENCIES = {"dev": ["pytest"]}

if sys.version_info < REQUIRED_PYTHON_VERSION:
    sys.exit("Python >= 3.0 is required. Your version:\n" + sys.version)


class PyTest(TestCommand):
    """
    Use pytest to run tests
    """

    user_options = [("pytest-args=", "a", "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name=NAME,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=URL,
    version=VERSION,
    author=AUTHOR,
    packages=PACKAGES,
    include_package_data=True,
    install_requires=INSTALL_DEPENDENCIES,
    setup_requires=SETUP_DEPENDENCIES,
    tests_require=TEST_DEPENDENCIES,
    extras_require=EXTRA_DEPENDENCIES,
    cmdclass={"test": PyTest},
)
